@@ -6,14 +6,16 @@
 # copied, modified, or distributed except according to those terms.
 
 import enum
+import importlib.metadata
 import logging
 import platform
 import time
-import warnings
 from abc import ABC, abstractmethod
+from collections.abc import Set
 from contextlib import contextmanager
+from importlib.metadata import PackageNotFoundError
 from io import BytesIO
-from typing import Any, Callable, Iterator, List, Optional
+from typing import TYPE_CHECKING, Any, Callable, Iterator, List, Optional, Union
 
 from nitrokey._helpers import Retries
 from nitrokey.nk3 import NK3, NK3Bootloader
@@ -27,40 +29,125 @@ from nitrokey.trussed._bootloader import (
 from nitrokey.trussed._bootloader.lpc55_upload.mboot.exceptions import (
     McuBootConnectionError,
 )
-from nitrokey.trussed.admin_app import BootMode
+from nitrokey.trussed.admin_app import BootMode, Status
+from nitrokey.trussed.admin_app import Variant as AdminAppVariant
 from nitrokey.updates import Asset, Release
 
+if TYPE_CHECKING:
+    import typing_extensions
+
 logger = logging.getLogger(__name__)
 
 
 @enum.unique
-class UpdatePath(enum.Enum):
-    default = enum.auto()
-    nRF_IFS_Migration_v1_3 = enum.auto()
+class Warning(enum.Enum):
+    """
+    A warning that can occur during a firmware update.
+
+    By default, these warnings abort the firmware update.  This enum can be used to select types
+    of warnings that should be ignored and not cause the firmware update to fail.
+    """
+
+    IFS_MIGRATION_V2 = "ifs-migration-v2"
+    MISSING_STATUS = "missing-status"
+    SDK_VERSION = "sdk-version"
+    UPDATE_FROM_BOOTLOADER = "update-from-bootloader"
+
+    @property
+    def message(self) -> str:
+        if self == Warning.IFS_MIGRATION_V2:
+            return (
+                "Not enough space on the internal filesystem to perform the firmware"
+                " update. See the release notes for more information:"
+                " https://github.com/Nitrokey/nitrokey-3-firmware/releases/tag/v1.8.2-test.20250312"
+            )
+        if self == Warning.MISSING_STATUS:
+            return (
+                "Could not determine the device state as the current firmware is too old."
+                " Please update to firmware version v1.3.1 first."
+            )
+        if self == Warning.SDK_VERSION:
+            return (
+                "Your Nitrokey SDK version is outdated.  Please update this program to the latest"
+                " version and try again."
+            )
+        if self == Warning.UPDATE_FROM_BOOTLOADER:
+            return (
+                "The current state of the device cannot be checked as it is already in bootloader"
+                " mode. Please review the release notes at:"
+                " https://github.com/Nitrokey/nitrokey-3-firmware/releases"
+            )
+
+        if TYPE_CHECKING:
+            typing_extensions.assert_never(self)
+
+        return self.value
 
     @classmethod
-    def create(
-        cls, variant: Optional[Variant], current: Optional[Version], new: Version
-    ) -> "UpdatePath":
+    def from_str(cls, s: str) -> "Warning":
+        for w in cls:
+            if w.value == s:
+                return w
+        raise ValueError(f"Unexpected update warning id: {s}")
+
+
+@enum.unique
+class _Migration(enum.Enum):
+    # IFS migration to use journaling on the NRF52 introduced in v1.3.0
+    NRF_IFS_MIGRATION = enum.auto()
+    # IFS migration to filesystem layout 2 in v1.8.2 (FIDO2 RK migration)
+    IFS_MIGRATION_V2 = enum.auto()
+
+    @classmethod
+    def get(
+        cls,
+        variant: Union[Variant, AdminAppVariant],
+        current: Optional[Version],
+        new: Version,
+    ) -> frozenset["_Migration"]:
+        if isinstance(variant, AdminAppVariant):
+            if variant == AdminAppVariant.USBIP:
+                raise ValueError("Cannot perform firmware update for USBIP runner")
+            elif variant == AdminAppVariant.LPC55:
+                variant = Variant.LPC55
+            elif variant == AdminAppVariant.NRF52:
+                variant = Variant.NRF52
+            else:
+                if TYPE_CHECKING:
+                    typing_extensions.assert_never(variant)
+
+                raise ValueError(f"Unsupported device variant: {variant}")
+
+        migrations = set()
+
         if variant == Variant.NRF52:
             if (
                 current is None
                 or current <= Version(1, 2, 2)
                 and new >= Version(1, 3, 0)
             ):
-                return cls.nRF_IFS_Migration_v1_3
-        return cls.default
+                migrations.add(cls.NRF_IFS_MIGRATION)
+
+        ifs_migration_v2 = Version(1, 8, 2)
+        if (
+            current is not None
+            and current < ifs_migration_v2
+            and new >= ifs_migration_v2
+        ):
+            migrations.add(cls.IFS_MIGRATION_V2)
+
+        return frozenset(migrations)
 
 
 def get_firmware_update(release: Release) -> Asset:
     return release.require_asset(Model.NK3.firmware_pattern)
 
 
-def get_extra_information(upath: UpdatePath) -> List[str]:
+def _get_extra_information(migrations: Set[_Migration]) -> List[str]:
     """Return additional information for the device after update based on update-path"""
 
     out = []
-    if upath == UpdatePath.nRF_IFS_Migration_v1_3:
+    if _Migration.NRF_IFS_MIGRATION in migrations:
         out += [
             "",
             "During this update process the internal filesystem will be migrated!",
@@ -71,11 +158,11 @@ def get_extra_information(upath: UpdatePath) -> List[str]:
     return out
 
 
-def get_finalization_wait_retries(upath: UpdatePath) -> int:
+def _get_finalization_wait_retries(migrations: Set[_Migration]) -> int:
     """Return number of retries to wait for the device after update based on update-path"""
 
     out = 60
-    if upath == UpdatePath.nRF_IFS_Migration_v1_3:
+    if _Migration.NRF_IFS_MIGRATION in migrations:
         # max time 150secs == 300 retries
         out = 500
     return out
@@ -86,6 +173,14 @@ class UpdateUi(ABC):
     def error(self, *msgs: Any) -> Exception:
         pass
 
+    @abstractmethod
+    def show_warning(self, warning: Warning) -> None:
+        pass
+
+    @abstractmethod
+    def raise_warning(self, warning: Warning) -> Exception:
+        pass
+
     @abstractmethod
     def abort(self, *msgs: Any) -> Exception:
         pass
@@ -120,12 +215,6 @@ class UpdateUi(ABC):
     def confirm_update_same_version(self, version: Version) -> None:
         pass
 
-    def request_repeated_update(self) -> Optional[Exception]:
-        warnings.warn(
-            "UpdateUi.request_repeated_update is no longer needed", DeprecationWarning
-        )
-        return None
-
     @abstractmethod
     def pre_bootloader_hint(self) -> None:
         pass
@@ -158,10 +247,18 @@ class Updater:
         await_device: Callable[
             [Optional[int], Optional[Callable[[int, int], None]]], NK3
         ],
+        ignore_warnings: Set[Warning] = frozenset(),
     ) -> None:
         self.ui = ui
         self.await_bootloader = await_bootloader
         self.await_device = await_device
+        self.ignore_warnings = ignore_warnings
+
+    def _trigger_warning(self, warning: Warning) -> None:
+        if warning in self.ignore_warnings:
+            self.ui.show_warning(warning)
+        else:
+            raise self.ui.raise_warning(warning)
 
     def update(
         self,
@@ -170,25 +267,39 @@ class Updater:
         update_version: Optional[str],
         ignore_pynitrokey_version: bool = False,
     ) -> Version:
-        current_version = device.admin.version() if isinstance(device, NK3) else None
+        update_from_bootloader = False
+        current_version = None
+        status = None
+        if isinstance(device, NK3Bootloader):
+            update_from_bootloader = True
+            self._trigger_warning(Warning.UPDATE_FROM_BOOTLOADER)
+        elif isinstance(device, NK3):
+            current_version = device.admin.version()
+            status = device.admin.status()
+        else:
+            raise self.ui.error(f"Unexpected Trussed device: {device}")
+
         logger.info(f"Firmware version before update: {current_version or ''}")
         container = self._prepare_update(image, update_version, current_version)
 
-        if container.pynitrokey:
-            # this is the version of pynitrokey when we moved to the SDK
-            pynitrokey_version = Version.from_str("0.4.49")
-            if container.pynitrokey > pynitrokey_version:
-                if ignore_pynitrokey_version:
-                    self.ui.confirm_pynitrokey_version(
-                        current=pynitrokey_version, required=container.pynitrokey
-                    )
-                else:
-                    raise self.ui.abort_pynitrokey_version(
-                        current=pynitrokey_version, required=container.pynitrokey
-                    )
+        if not update_from_bootloader:
+            if status is None and container.version > Version.from_str("1.3.1"):
+                self._trigger_warning(Warning.MISSING_STATUS)
+
+        self._check_minimum_version(container, ignore_pynitrokey_version)
 
         self.ui.confirm_update(current_version, container.version)
 
+        migrations = None
+        if status is not None and status.variant is not None:
+            migrations = self._check_migrations(
+                status.variant, current_version, container.version, status
+            )
+        elif isinstance(device, NK3Bootloader):
+            migrations = self._check_migrations(
+                device.variant, current_version, container.version, status
+            )
+
         with self._get_bootloader(device) as bootloader:
             if bootloader.variant not in container.images:
                 raise self.ui.error(
@@ -205,15 +316,14 @@ class Updater:
             except Exception as e:
                 raise self.ui.error("Failed to validate firmware image", e)
 
-            update_path = UpdatePath.create(
-                bootloader.variant, current_version, container.version
-            )
-            txt = get_extra_information(update_path)
-            self.ui.confirm_extra_information(txt)
+            if migrations is None:
+                migrations = self._check_migrations(
+                    bootloader.variant, current_version, container.version, status
+                )
 
             self._perform_update(bootloader, container)
 
-        wait_retries = get_finalization_wait_retries(update_path)
+        wait_retries = _get_finalization_wait_retries(migrations)
         with self.ui.finalization_progress_bar() as callback:
             with self.await_device(wait_retries, callback) as device:
                 version = device.admin.version()
@@ -295,6 +405,36 @@ class Updater:
 
         return container
 
+    def _check_minimum_version(
+        self, container: FirmwareContainer, ignore_pynitrokey_version: bool
+    ) -> None:
+        if container.sdk:
+            try:
+                sdk_version = Version.from_str(importlib.metadata.version("nitrokey"))
+            except PackageNotFoundError:
+                raise self.ui.error("Failed to determine the Nitrokey SDK version")
+
+            if container.sdk > sdk_version:
+                logger.warning(
+                    f"Minimum SDK version required for update is {container.sdk} (current version: {sdk_version})"
+                )
+                self._trigger_warning(Warning.SDK_VERSION)
+        elif container.pynitrokey:
+            # The minimum pynitrokey version has been replaced by the minimum SDK version, so we
+            # only check it if there is no minimum SDK version set.
+
+            # this is the version of pynitrokey when we moved to the SDK
+            pynitrokey_version = Version.from_str("0.4.49")
+            if container.pynitrokey > pynitrokey_version:
+                if ignore_pynitrokey_version:
+                    self.ui.confirm_pynitrokey_version(
+                        current=pynitrokey_version, required=container.pynitrokey
+                    )
+                else:
+                    raise self.ui.abort_pynitrokey_version(
+                        current=pynitrokey_version, required=container.pynitrokey
+                    )
+
     def _validate_version(
         self,
         current_version: Optional[Version],
@@ -351,6 +491,29 @@ class Updater:
         else:
             raise self.ui.error(f"Unexpected Nitrokey 3 device: {device}")
 
+    def _check_migrations(
+        self,
+        variant: Union[Variant, AdminAppVariant],
+        current_version: Optional[Version],
+        new_version: Version,
+        status: Optional[Status],
+    ) -> frozenset["_Migration"]:
+        try:
+            migrations = _Migration.get(
+                variant=variant, current=current_version, new=new_version
+            )
+        except ValueError as e:
+            raise self.ui.error(str(e))
+
+        txt = _get_extra_information(migrations)
+        self.ui.confirm_extra_information(txt)
+
+        if _Migration.IFS_MIGRATION_V2 in migrations:
+            if status and status.ifs_blocks is not None and status.ifs_blocks < 5:
+                self._trigger_warning(Warning.IFS_MIGRATION_V2)
+
+        return migrations
+
     def _perform_update(
         self, device: NK3Bootloader, container: FirmwareContainer
     ) -> None:
